# -*- coding: utf-8 -*-
# copyright 2014 UNLISH, all rights reserved.
# contact http://www.unlish.com/ -- mailto:christophe@unlish.com
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""cubicweb-wsme views/forms/actions/components for web ui"""


from cubicweb.web.views.urlpublishing import URLPathEvaluator, PathDontMatch


class RestPathEvaluator(URLPathEvaluator):
    # Priority is set to be have more weight than the default RestPathEvaluator
    priority = 1

    def evaluate_path(self, req, parts):
        if not (0 < len(parts) < 6):
            raise PathDontMatch()
        try:
            etype = self.vreg.case_insensitive_etypes[parts.pop(0).lower()]
        except KeyError:
            raise PathDontMatch()

        accept = set(req.parse_accept_header('Accept'))

        if not accept.intersection(
                ('application/json', 'text/javascript', 'text/xml')):
            raise PathDontMatch()

        # First element must be an etype
        cls = self.vreg['etypes'].etype_class(etype)
        rset = None

        # Get an optional value, ie an entity identifier
        if parts:
            value = parts.pop(0)
            if value in cls.e_schema.subjrels:
                attrname = value
                value = parts.pop(0)
            else:
                attrname = cls.cw_rest_attr_info()[0]
            value = req.url_unquote(value)
            # XXX check the rschema instead
            if attrname == 'eid':
                value = int(value)
            rset = req.find(etype, **{attrname: value})
            if rset.rowcount != 1:
                raise PathDontMatch()

        # Get an optional rtype or function
        func = rtype = None
        if parts:
            part = parts.pop(0)
            if part.startswith('!'):
                func = part[1:]
            else:
                rtype = part
        else:
            rtype = None

        # Get an optional relation target id
        if parts:
            if not rtype:
                raise PathDontMatch()
            rtype_target = parts.pop(0)
        else:
            rtype_target = None

        method = parts.pop(0) if parts else req.http_method().lower()
        req.form['_ws_method'] = method
        req.form['_ws_etype'] = etype
        req.form['_ws_func'] = func
        req.form['_ws_rtype'] = rtype
        req.form['_ws_rtype_target'] = rtype_target
        return 'webservice', rset



from cubes.wsme.types import Base, wsattr, scan


class CWUser(Base):
    __etype__ = 'CWUser'

    __autoattr__ = True
    __autoexclude__ = Base.__autoexclude__ + ('in_state',)

    password = wsattr('upassword', writeonly=True)


def registration_callback(vreg):
    scan(vreg, __name__)



# copyright 2014 Christophe de Vienne, all rights reserved.
# contact http://www.unlish.com/ -- mailto:christophe@unlish.com
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""cubicweb-wsme tests
"""

import base64
import json
import urllib

import wsme

from cubicweb.devtools.testlib import CubicWebTC
from cubicweb.devtools.httptest import CubicWebServerTC
import cubicweb

from cubes.wsme.types import Base, wsattr
from cubes.wsme.controller import to_fetchtree, prefetch


class WSMETest(CubicWebTC):
    @classmethod
    def init_config(cls, config):
        super(WSMETest, cls).init_config(config)
        config.debugmode = True

    def test_types_final_rel(self):
        class ACWUser(Base):
            __etype__ = 'CWUser'
            login = wsattr('login', datatype=wsme.types.text)
            # XXX in real-life, the upassword would never be copied from an
            # entity to a ws type, only from a ws type to an entity (with
            # encryption at the same time)
            password = wsattr('upassword', datatype=wsme.types.text)

        ACWUser.reginit(self.vreg)

        with self.admin_access.repo_cnx() as cnx:
            u = cnx.find("CWUser", login=u"admin").one()
            u_eid = u.eid
            ws_u = ACWUser(u)

        self.assertEqual(u_eid, ws_u.eid)
        self.assertEqual(u"admin", ws_u.login)

        ws_u.login = u"anotherone"
        ws_u.password = 'somepassword'
        with self.admin_access.repo_cnx() as cnx:
            cls = self.vreg['etypes'].etype_class('CWUser')
            u = cls.cw_instantiate(cnx.execute, **ws_u.final_values())

            self.assertNotEqual(u_eid, u.eid)
            self.assertEqual(u"anotherone", ws_u.login)

    def test_types_final_rel_autoattr(self):
        class ACWUser(Base):
            __etype__ = 'CWUser'

            login = wsattr()
            password = wsattr('upassword')

        ACWUser.reginit(self.vreg)
        self.vreg.wsme_registry.finalize_init()

        with self.admin_access.repo_cnx() as cnx:
            u = cnx.find("CWUser", login=u"admin").one()
            u_eid = u.eid
            ws_u = ACWUser(u)

        self.assertEqual(u_eid, ws_u.eid)
        self.assertEqual(u"admin", ws_u.login)

    def test_types_relations(self):
        class ACWUser(Base):
            __etype__ = 'CWUser'

            login = wsattr('login', datatype=wsme.types.text)
            # XXX in real-life, the upassword would never be copied from an
            # entity to a ws type, only from a ws type to an entity (with
            # encryption at the same time)
            password = wsattr('upassword', datatype=wsme.types.text)

            in_group = wsattr('in_group', datatype=['ACWGroup'])

        class ACWGroup(Base):
            __etype__ = 'CWGroup'

            name = wsattr('name', datatype=wsme.types.text)
            users = wsattr('in_group', role='object', datatype=[ACWUser])

        ACWUser.reginit(self.vreg)
        ACWGroup.reginit(self.vreg)

        with self.admin_access.repo_cnx() as cnx:
            u = cnx.find("CWUser", login=u"admin").one()
            u_eid = u.eid
            g_eid = u.in_group[0].eid
            ws_u = ACWUser(u, fetch=['in_group'])
            ws_u2 = ACWUser(u, fetch=['in_group.name'])

        self.assertEqual(u_eid, ws_u.eid)
        self.assertEqual(u"admin", ws_u.login)

        self.assertEqual(g_eid, ws_u.in_group[0].eid)
        self.assertEqual(u"managers", ws_u2.in_group[0].name)

    def test_polymorphic_relations(self):
        State = self.vreg.wsme_registry.lookup('State')

        with self.admin_access.repo_cnx() as cnx:
            state = cnx.find('State', name='activated').one()
            tr_eid = state.allowed_transition[0].eid
            tr_created_by_eid = state.allowed_transition[0].created_by[0].eid
            ws_state = State(state, fetch=['allowed_transition.created_by'])

        self.assertEqual(tr_eid, ws_state.allowed_transition[0].eid)
        self.assertEqual('Transition', ws_state.allowed_transition[0].cw_etype)
        self.assertEqual(
            tr_created_by_eid,
            ws_state.allowed_transition[0].created_by.eid)

    def test_object_relations(self):
        CWGroup = self.vreg.wsme_registry.lookup('CWGroup')
        self.assertTrue(hasattr(CWGroup, 'in_group_object'))
        self.assertEqual('<in_group', CWGroup.in_group_object.name)


class WSMECRUDTest(CubicWebServerTC):
    @classmethod
    def init_config(cls, config):
        super(WSMECRUDTest, cls).init_config(config)
        config.debugmode = True

    def test_cwuser_get(self):
        url = 'cwuser?' + urllib.urlencode({'filter': '{"login": "anon"}'})
        res = self.web_request(
            url, headers={"Accept": "application/json"})
        if '<!DOCTYPE html>' in res.body:
            import html2text
            print html2text.html2text(res.body)
        data = json.loads(res.body)
        self.assertNotIn('faultstring', data)
        self.assertEqual(u"anon", data[0]['login'])

    def test_cwuser_get_keyonly(self):
        url = 'cwuser?' + urllib.urlencode({'keyonly': True})
        res = self.web_request(
            url, headers={"Accept": "application/json"})
        if '<!DOCTYPE html>' in res.body:
            import html2text
            print html2text.html2text(res.body)
        data = json.loads(res.body)
        self.assertNotIn('login', data[0])
        self.assertIn('modification_date', data[0])

    def test_cwuser_entity_get(self):
        url = 'cwuser/anon?fetch=in_group'
        res = self.web_request(
            url, headers={"Accept": "application/json"})
        if '<!DOCTYPE html>' in res.body:
            import html2text
            print html2text.html2text(res.body)
        data = json.loads(res.body)
        self.assertEqual(u"anon", data['login'])
        self.assertEqual(1, len(data['in_group']))
        self.assertIn('modification_date', data['in_group'][0])

    def test_cwuser_entity_get_fetch_typed_relation(self):
        with self.admin_access.repo_cnx() as cnx:
            cnx.find('CWGroup', name=u'managers').one().cw_set(
                created_by=cnx.find('CWUser', login=u'admin').one())
            cnx.commit()
        self.web_login()
        url = 'cwuser/admin?fetch=<created_by[CWGroup]'
        res = self.web_request(
            url, headers={"Accept": "application/json"})
        if '<!DOCTYPE html>' in res.body:
            import html2text
            print html2text.html2text(res.body)
        data = json.loads(res.body)
        self.assertEqual(200, res.status, data)
        self.assertEqual(u"admin", data['login'])
        self.assertEqual(1, len(data['<created_by']))
        self.assertIn('modification_date', data['<created_by'][0])

    def test_cwgroup_post(self):
        self.web_login()
        url = 'cwgroup'
        res = self.web_request(
            url, method='POST',
            body='{"name": "test"}',
            headers={"Accept": "application/json",
                     "Content-Type": "application/json"})
        if '<!DOCTYPE html>' in res.body:
            import html2text
            print html2text.html2text(res.body)
            assert False

        data = json.loads(res.body)
        self.assertEqual(200, res.status, data)

        self.assertEqual(u"test", data['name'])
        g_eid = data['eid']

        with self.admin_access.repo_cnx() as cnx:
            self.assertEqual(u"test", cnx.entity_from_eid(g_eid).name)

    def test_cwgroup_entity_put(self):
        with self.admin_access.repo_cnx() as cnx:
            g_eid = cnx.create_entity('CWGroup', name=u'test').eid
            cnx.commit()

        self.web_login()
        url = 'cwgroup/test'
        res = self.web_request(
            url, method='PUT',
            body='{"name": "newname"}',
            headers={"Accept": "application/json",
                     "Content-Type": "application/json"})
        if '<!DOCTYPE html>' in res.body:
            import html2text
            print html2text.html2text(res.body)
            assert False
        self.assertEqual(200, res.status)
        with self.admin_access.repo_cnx() as cnx:
            self.assertEqual(u"newname", cnx.entity_from_eid(g_eid).name)

    def test_cwgroup_entity_delete(self):
        with self.admin_access.repo_cnx() as cnx:
            cnx.create_entity('CWGroup', name=u'test').eid
            cnx.commit()

        self.web_login()
        url = 'cwgroup/test'
        res = self.web_request(
            url, method='DELETE', headers={"Accept": "application/json"})
        if '<!DOCTYPE html>' in res.body:
            import html2text
            print html2text.html2text(res.body)
            assert False
        self.assertEqual(204, res.status)

    def test_multi_create(self):
        self.web_login()

        res = self.web_request(
            'cwuser?fetch=in_group', method='POST',
            headers={"Accept": "application/json",
                     "Content-Type": "application/json"},
            body=json.dumps({
                'login': "AUser", 'password': "APassword",
                'in_group': [{
                    'name': 'ANewGroup'
                }]
            }),
        )

        data = json.loads(res.body)
        self.assertNotIn('faultcode', data)

        self.assertEqual(200, res.status)

        g_eid = data['in_group'][0]['eid']
        with self.admin_access.repo_cnx() as cnx:
            self.assertEqual(u"ANewGroup", cnx.entity_from_eid(g_eid).name)

    def test_multi_update(self):
        with self.admin_access.repo_cnx() as cnx:
            g_eid = cnx.find('CWGroup', name=u'managers').one().eid

        self.web_login()

        res = self.web_request(
            'cwuser/admin?fetch=in_group', method='PUT',
            headers={"Accept": "application/json",
                     "Content-Type": "application/json"},
            body=json.dumps({
                'surname': 'Admin',
                'in_group': [{
                    'eid': g_eid,
                    'name': 'newname'
                }]
            }),
        )

        data = json.loads(res.body)
        self.assertNotIn('faultcode', data)

        self.assertEqual(200, res.status)

        self.assertEqual(u'Admin', data['surname'])
        self.assertEqual(g_eid, data['in_group'][0]['eid'])

        with self.admin_access.repo_cnx() as cnx:
            self.assertEqual(u"newname", cnx.entity_from_eid(g_eid).name)

    def test_update_and_create(self):
        with self.admin_access.repo_cnx() as cnx:
            g_eid = cnx.find('CWGroup', name=u'managers').one().eid

        self.web_login()

        res = self.web_request(
            'cwuser/admin?fetch=in_group', method='PUT',
            headers={"Accept": "application/json",
                     "Content-Type": "application/json"},
            body=json.dumps({
                'surname': "AdminName",
                'in_group': [{
                    'name': 'ANewGroup'
                }, {
                    'eid': g_eid,
                }]
            }),
        )

        data = json.loads(res.body)
        self.assertNotIn('faultcode', data)

        self.assertEqual(200, res.status)

        g_eid = data['in_group'][0]['eid']
        with self.admin_access.repo_cnx() as cnx:
            self.assertEqual(u"ANewGroup", cnx.entity_from_eid(g_eid).name)

    def test_create_and_update(self):
        with self.admin_access.repo_cnx() as cnx:
            g_eid = cnx.find('CWGroup', name=u'managers').one().eid

        self.web_login()

        res = self.web_request(
            'cwuser', method='POST',
            headers={"Accept": "application/json",
                     "Content-Type": "application/json"},
            body=json.dumps({
                'login': "AUser", 'password': "APassword",
                'in_group': [{
                    'eid': g_eid,
                    'name': 'NewName'
                }]
            }),
        )

        data = json.loads(res.body)
        self.assertNotIn('faultcode', data)

        self.assertEqual(200, res.status)

        with self.admin_access.repo_cnx() as cnx:
            self.assertEqual(u"NewName", cnx.entity_from_eid(g_eid).name)

    def test_get_relation(self):
        self.web_login()

        res = self.web_request(
            'cwgroup/managers/<in_group?keyonly=1&limit=2', method='GET',
            headers={"Accept": "application/json",
                     "Content-Type": "application/json"})

        data = json.loads(res.body)
        self.assertNotIn('faultcode', data)

        self.assertEqual(200, res.status)
        self.assertEqual(1, len(data))
        self.assertEqual('CWUser', data[0]['cw_etype'])

    def test_add_relation(self):
        with self.admin_access.repo_cnx() as cnx:
            u_eid = cnx.find('CWUser', login=u'anon').one().eid

        self.web_login()

        res = self.web_request(
            'cwgroup/managers/<in_group',
            method='POST',
            body=str(u_eid),
            headers={"Accept": "application/json",
                     "Content-Type": "application/json"})

        self.assertEqual(200, res.status)

        with self.admin_access.repo_cnx() as cnx:
            g = cnx.find('CWGroup', name=u'managers').one()
            u = cnx.find('CWUser', login=u'anon').one()

            self.assertIn(g, u.in_group)

    def test_del_relation(self):
        with self.admin_access.repo_cnx() as cnx:
            u = cnx.find('CWUser', login=u'anon').one()
            g = cnx.find('CWGroup', name=u'managers').one()
            u.cw_set(in_group=g)
            g_eid = g.eid
            cnx.commit()

        self.web_login()

        res = self.web_request(
            'cwuser/anon/in_group/%s' % (g_eid),
            method='DELETE',
            headers={"Accept": "application/json"})

        self.assertEqual(204, res.status)

        with self.admin_access.repo_cnx() as cnx:
            u = cnx.find('CWUser', login=u'anon').one()

            self.assertEqual(1, len(u.in_group))

    def test_get_binary(self):
        with self.admin_access.repo_cnx() as cnx:
            f = cnx.create_entity(
                'File', data=cubicweb.Binary('test'),
                data_name=u'filename',
                data_format=u'text/plain')
            cnx.commit()
            eid = f.eid

        self.web_login()

        res = self.web_request(
            'file/{}'.format(eid),
            headers={"Accept": "application/json"})

        data = json.loads(res.body)
        self.assertNotIn('faultcode', data)
        self.assertEqual(200, res.status)

        self.assertEqual(data['data'], base64.encodestring('test'))

    def test_set_binary(self):
        with self.admin_access.repo_cnx() as cnx:
            f = cnx.create_entity(
                'File', data=cubicweb.Binary('test'),
                data_name=u'filename',
                data_format=u'text/plain')
            cnx.commit()
            eid = f.eid

        self.web_login()

        res = self.web_request(
            'file/{}'.format(eid),
            method='PUT',
            body=json.dumps({"data": base64.encodestring('newcontent')}),
            headers={
                "Accept": "application/json",
                "Content-Type": "application/json"})

        data = json.loads(res.body)
        self.assertNotIn('faultcode', data)
        self.assertEqual(200, res.status)

        self.assertEqual(data['data'], base64.encodestring('newcontent'))

        with self.admin_access.repo_cnx() as cnx:
            f = cnx.entity_from_eid(eid)
            self.assertEqual(f.data.getvalue(), 'newcontent')


class WSMEQueryTest(CubicWebServerTC):
    @classmethod
    def init_config(cls, config):
        super(WSMEQueryTest, cls).init_config(config)
        config.debugmode = True

    def test_orderby(self):
        self.web_login()

        res = self.web_request(
            'cwuser?orderby=login',
            method='GET',
            headers={"Accept": "application/json"})

        data = json.loads(res.body)
        self.assertNotIn('faultcode', data)
        self.assertEqual(
            ['admin', 'anon'],
            [d['login'] for d in data])

        res = self.web_request(
            'cwuser?orderby=-login',
            method='GET',
            headers={"Accept": "application/json"})

        data = json.loads(res.body)
        self.assertNotIn('faultcode', data)
        self.assertEqual(
            ['anon', 'admin'],
            [d['login'] for d in data])

    def test_limit(self):
        self.web_login()

        res = self.web_request(
            'cwuser?limit=1',
            method='GET',
            headers={"Accept": "application/json"})

        data = json.loads(res.body)
        self.assertEqual(1, len(data))

    def test_offset(self):
        self.web_login()

        # we need a limit because sqlite seems not to like an offset without a
        # limit
        res = self.web_request(
            'cwuser?limit=1&offset=2',
            method='GET',
            headers={"Accept": "application/json"})

        data = json.loads(res.body)
        self.assertNotIn('faultcode', data)
        self.assertEqual(0, len(data))

    def test_filter_on_one_attribute(self):
        self.web_login()

        res = self.web_request(
            'cwuser?' + urllib.urlencode({
                'filter': json.dumps({
                    'login': 'admin'
                })
            }),
            method='GET',
            headers={"Accept": "application/json"}
        )

        data = json.loads(res.body)
        self.assertNotIn('faultcode', data)

        self.assertEqual(1, len(data))
        self.assertEqual(u'admin', data[0]['login'])

    def test_filter_on_attribute_and_relation(self):
        self.web_login()

        res = self.web_request(
            'cwuser?' + urllib.urlencode({
                'filter': json.dumps({
                    'login': {'$ne': 'anon'},
                    'in_group': {
                        'name': 'managers'
                    }
                })
            }),
            method='GET',
            headers={"Accept": "application/json"}
        )

        data = json.loads(res.body)
        self.assertNotIn('faultcode', data)

        self.assertEqual(1, len(data))
        self.assertEqual(u'admin', data[0]['login'])

    def test_unset_relation(self):
        with self.admin_access.repo_cnx() as cnx:
            g_eid = cnx.create_entity('CWGroup', name=u'test').eid
            cnx.find('CWUser', login=u'admin').one().cw_set(in_group=g_eid)
            cnx.commit()

        self.web_login()

        url = 'cwgroup/test'
        res = self.web_request(
            url, method='PUT',
            body='{"<in_group": null}',
            headers={"Accept": "application/json",
                     "Content-Type": "application/json"})
        if '<!DOCTYPE html>' in res.body:
            import html2text
            print html2text.html2text(res.body)
            assert False
        self.assertEqual(200, res.status)
        with self.admin_access.repo_cnx() as cnx:
            self.assertEqual(
                (),
                cnx.entity_from_eid(g_eid).reverse_in_group)

    def test_call_function(self):
        with self.admin_access.repo_cnx() as cnx:
            u = cnx.create_entity(
                'CWUser', login=u"newuser", upassword=u"xx",
                in_group=cnx.find('CWGroup', name='users').one())
            cnx.commit()
            self.assertEqual(u.cw_adapt_to('IWorkflowable').state, 'activated')
            u.cw_clear_relation_cache()
        self.web_login()

        res = self.web_request(
            'cwuser/newuser/!IWorkflowable.fire_transition', method='POST',
            body='{"tr": "deactivate", "comment": "from the tests"}',
            headers={"Accept": "application/json",
                     "Content-Type": "application/json"})

        data = json.loads(res.body)
        self.assertNotIn('faultcode', data)

        self.assertIn('eid', data)
        self.assertIn('modification_date', data)

        with self.admin_access.repo_cnx() as cnx:
            u = cnx.entity_from_eid(u.eid)
            self.assertEqual(
                u.cw_adapt_to('IWorkflowable').state, 'deactivated')

    def test_to_fetchtree(self):
        self.assertEqual(
            to_fetchtree(['in_group.']),
            {'in_group': {'': {}}})

        self.assertEqual(
            to_fetchtree(['in_group']),
            {'in_group': {}})

        self.assertEqual(
            to_fetchtree(['in_group.name']),
            {'in_group': {'name': {}}})

        self.assertEqual(
            to_fetchtree(['in_group.name', 'in_group.owned_by']),
            {'in_group': {'name': {}, 'owned_by': {}}})

        self.assertEqual(
            to_fetchtree(['in_group.', 'in_group.owned_by']),
            {'in_group': {'': {}, 'owned_by': {}}})

    def test_prefetch(self):
        with self.admin_access.repo_cnx() as cnx:
            users = list(cnx.execute(
                'CWUser U ORDERBY L WHERE U login L').entities())
            admin, anon = users
            prefetch(cnx, users, {'in_group': {'name': {}}})
            self.assertIsNotNone(
                admin.cw_relation_cached('in_group', 'subject'))
            self.assertIn('name', admin.in_group[0].cw_attr_cache)
            self.assertNotIn('creation_date', admin.in_group[0].cw_attr_cache)



import datetime
import decimal
import weakref

from cubicweb import cwvreg

from cubes.wsme.types import Base, Any, JsonData, binary
from cubes.wsme.controller import WSCRUDController

import wsme.types


class WSMERegistry(wsme.types.Registry):
    """The wsme types registry.
    To be populated before type solving"""

    @staticmethod
    def get(vreg):
        instance = getattr(vreg, '_wsme_registry', None)
        if instance is None:
            vreg._wsme_registry = instance = WSMERegistry(vreg)
            Any.reginit(vreg)
            instance.register(JsonData)

        return instance

    def __init__(self, vreg):
        super(WSMERegistry, self).__init__()

        self.vreg = weakref.proxy(vreg)

        self.cwtypes = {
            'Boolean': bool,
            'Int': long,
            'BigInt': long,
            'Float': float,
            'Boolean': bool,
            'Decimal': decimal.Decimal,
            'Date': datetime.date,
            'Time': datetime.time,
            'Datetime': datetime.datetime,
            'TZDatetime': datetime.datetime,
            'String': wsme.types.text,
            'Password': wsme.types.text,
            'Bytes': binary
        }

    def finalize_init(self):
        for class_ in self.complex_types:
            if issubclass(class_, Base):
                class_.finalize_init()

    def guess_datatype(self, etype):
        return self.cwtypes.get(etype)

    def register(self, class_):
        class_ = super(WSMERegistry, self).register(class_)
        etype = getattr(class_, '__etype__', None)

        if etype is not None:
            self.cwtypes[etype] = class_

        return class_


class ControllerRegistry(cwvreg.CWRegistry):
    def initialization_completed(self):
        etypes = set()
        # make sure all etypes have a corresponding wstype
        for eschema in self.vreg.schema.entities():
            if not eschema.final:
                etypes.add(eschema.type)

            wstype = self.vreg.wsme_registry.guess_datatype(eschema.type)
            if wstype is None:
                new_class = type(eschema.type, (Base,), {
                    '__etype__': eschema.type,
                    '__autoattr__': True
                })
                new_class.reginit(self.vreg)

        self.vreg.wsme_registry.finalize_init()

        # make sure there is a wscontroller for each etype
        controllers = {
            ctrl.__cwetype__.__etype__: ctrl
            for ctrl in (
                self.vreg['controllers']['webservice']
                if 'webservice' in self.vreg['controllers']
                else ())
        }

        for etype in etypes.difference(controllers.keys()):
            wstype = self.vreg.wsme_registry.guess_datatype(etype)
            new_ctrl = type(etype + 'Controller', (WSCRUDController,), {
                '__cwetype__': wstype
            })
            self.vreg.register(new_ctrl)

        super(ControllerRegistry, self).initialization_completed()


cwvreg.CWRegistryStore.REGISTRY_FACTORY['controllers'] = ControllerRegistry
cwvreg.CWRegistryStore.wsme_registry = property(WSMERegistry.get)



from cubicweb.predicates import Predicate


class match_ws_etype(Predicate):
    def __init__(self, etype):
        self.etype = etype

    def __call__(self, cls, req, **kwargs):
        return 1 if req.form.get('_ws_etype') == self.etype else 0



# -*- coding: utf-8 -*-
# copyright 2014 UNLISH, all rights reserved.
# contact http://www.unlish.com/ -- mailto:christophe@unlish.com
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""
Types
"""

import base64
import inspect
import itertools
import json
import re

import six
import wsme
import wsme.types
import wsme.rest.args
import wsme.rest.json

from cubicweb import Unauthorized, Binary


class PassThroughType(object):
    """Special webservice type that transmit a value without doing anything

    It is used in wsme signature for the 'entity' argument which is passed by
    the controller itself and should not be transtyped in any way."""

    @classmethod
    def validate(self, value):
        return value


@wsme.rest.args.from_param.when_object(PassThroughType)
def any_from_param(datatype, value):
    return value


class JsonDataType(wsme.types.UserType):
    name = 'json'
    basetype = wsme.types.text

    def frombasetype(self, value):
        if value is None:
            return value
        return json.loads(value)

    def tobasetype(self, value):
        return json.dumps(value)

    @classmethod
    def validate(self, value):
        return value

#: User type that carry json encoded arbitrary data.
JsonData = JsonDataType()


@wsme.rest.json.fromjson.when_object(JsonData)
def jsondata_fromjson(datatype, value):
    return value


class wsattr(wsme.types.wsattr):
    """Cubicweb-specific version of :class:`wsme.types.wsattr`

    To be used on a :class:`Base` class.
    All the attributes are optionnal and can be deduced from the :attr:`rtype`.

    .. attribute:: rtype

        The corresponding relation in the model

    .. attribute:: role

        The role of the parent class in the relation

    .. attribute:: etype

        The entity type on the other side of the relation

    .. attribute:: datatype

        The webservice type
    """
    def __init__(
            self, rtype=None, role='subject', etype=None, datatype=None,
            writeonly=False, **kw):
        super(wsattr, self).__init__(datatype, **kw)
        self.rtype = rtype
        self.role = role
        self.etype = etype
        self.writeonly = writeonly

    def reginit(self, vreg, class_):
        if self.rtype is None:
            self.rtype = self.name
        rschema = vreg.schema.rschema(self.rtype)
        self.isfinal = rschema.final
        self.inlined = rschema.inlined

        etypes = rschema.targets(
            None if class_.__etype__ == 'Any' else class_.__etype__, self.role)
        if self.role == 'subject':
            self.rname = self.rtype
        elif self.role == 'object':
            self.rname = 'reverse_' + self.rtype

        if self.etype is None:
            # Autodetect the etype
            if len(etypes) != 1:
                self.etype = 'Any'
            else:
                self.etype = etypes[0].type

        else:
            # check that the etype is valid
            if self.etype == 'Any':
                if len(etypes) == 1:
                    raise ValueError(
                        'Any should not be used on non-polymorphic relations')
            else:
                if self.etype not in etypes:
                    raise ValueError(
                        'Wrong etype %s for relation %s %s' %
                        (self.etype, self.rtype, self.role))

    def finalize_init(self, class_):
        if self.datatype is not None and self.datatype != [None]:
            return

        rschema = class_.__registry__.vreg.schema.rschema(self.rtype)
        datatype = class_.__registry__.guess_datatype(self.etype)

        if class_.__etype__ == 'Any' or self.etype == 'Any':
            needarray = False
            for rdef in rschema.rdefs.values():
                if rdef.role_cardinality(self.role) in '+*':
                    needarray = True
                    break
        else:
            rdef = rschema.role_rdef(class_.__etype__, self.etype, self.role)
            needarray = rdef.role_cardinality(self.role) in '+*'
        if needarray:
            datatype = wsme.types.ArrayType(datatype)
        self.datatype = datatype


def iswsattr(obj):
    """returns True if an object is a :class:`wsattr`"""
    return isinstance(obj, wsattr)


class BaseMeta(wsme.types.BaseMeta):
    """
    Metaclass for :class:`Base`

    Bypass the wsme.types.BaseMeta __new__ and __init__
    because they set a default __registry__
    """

    def __new__(cls, name, bases, dct):
        return type.__new__(cls, name, bases, dct)

    def __init__(cls, name, bases, dct):
        pass


class Base(six.with_metaclass(BaseMeta, wsme.types.Base)):
    """Base class for a complex type that map an entity type
    """
    #: Entity eid
    eid = long

    __autoexclude__ = (
        'is', 'identity', 'cw_source',
        'is_instance_of', 'has_text')

    def __init__(self, entity=None, keyonly=False, fetch=()):
        if entity:
            self.from_entity(entity, keyonly, fetch)

    @classmethod
    def reginit(cls, vreg):
        """ Register the class

        Use the informations in the registry, and most notably the schema, to
        initialize the  attributes.
        """
        if getattr(cls, '__autoattr__', False):
            eschema = vreg.schema.eschema(cls.__etype__)

            existing_rels = set(
                (attr.rtype, attr.role)
                for name, attr in inspect.getmembers(cls, iswsattr))

            if cls.__autoattr__ is True:
                rels = itertools.chain(
                    ((rel, 'subject', rel.type) for rel in
                        eschema.subject_relations()),
                    ((rel, 'object', '<' + rel.type) for rel in
                        eschema.object_relations()))
            else:
                rels = itertools.chain(
                    ((eschema.subjrels[name], 'subject', name)
                     for name in cls.__autoattr__
                     if not name.startswith('<')),
                    ((eschema.objrels[name[1:]], 'object', name)
                     for name in cls.__autoattr__
                     if name.startswith('<')))
            rels = (
                (rel, role, name)
                for rel, role, name in rels
                if name not in cls.__autoexclude__
                and (rel.type, role) not in existing_rels
            )
            for rel, role, name in rels:
                key = rel.type
                if role == 'object':
                    key += '_object'
                setattr(
                    cls, key,
                    wsattr(rel.type, role=role, name=name))

        vreg.wsme_registry.register(cls)

        for attr in cls._wsme_attributes:
            if isinstance(attr, wsattr):
                attr.reginit(vreg, cls)

    @classmethod
    def finalize_init(cls):
        """ Finalize the class initialization.

        This last step resolve types in the underlying attributes.
        """
        for attr in cls._wsme_attributes:
            if isinstance(attr, wsattr):
                attr.finalize_init(cls)

    _rtype_re = re.compile(u'(?P<rtype>[^.[]+)\[(?P<etypes>[^\]]+)\]')

    def from_entity(self, entity, keyonly=False, fetch=()):
        """ Load values from an entity

        :param entity: the entity
        :param keyonly: if `True`, only the .eid and .modfication_date will be
                        loaded. The result can be used as "timestamped
                        reference".
        :param fetch: a list of relations to eager load. Unless specified, all
                      the '1' or '?' relation targets will be loaded as
                      'keyonly', and the '*' or '+' relations will not be
                      loaded at all.
        """
        self.eid = entity.eid
        self.modification_date = entity.modification_date

        if keyonly:
            # Append the modification_date
            return

        typed_fetch = {
            m.group('rtype'): m.group('etypes').split(',')
            for m in (
                self._rtype_re.match(f) for f in fetch) if m is not None}

        for attr in self._wsme_attributes:
            if not isinstance(attr, wsattr):
                continue

            if attr.writeonly:
                continue

            if attr.isfinal:
                subfetch = None
            else:
                prefix = attr.name + '.'
                prefixlen = len(prefix)
                subfetch = [
                    x[prefixlen:] for x in fetch if x.startswith(prefix)
                ]

            if wsme.types.isarray(attr.datatype):
                if not attr.name in fetch \
                        and not attr.name in typed_fetch \
                        and not subfetch:
                    continue

            if attr.name in typed_fetch:
                value = entity.related(
                    attr.rtype, role=attr.role, entities=True,
                    targettypes=typed_fetch[attr.name])
            else:
                try:
                    value = getattr(entity, attr.rname)
                except Unauthorized:
                    # XXX set value = Forbidden ?
                    continue

            if not attr.isfinal:
                attr_keyonly = not subfetch
                if wsme.types.isarray(attr.datatype):
                    if value:
                        value = [
                            attr.datatype.item_type(
                                o,
                                keyonly=attr_keyonly,
                                fetch=subfetch)
                            for o in value
                        ]
                    else:
                        value = []
                else:
                    if value:
                        value = attr.datatype(
                            value[0],
                            keyonly=attr_keyonly,
                            fetch=subfetch
                        )
                    else:
                        value = None

            attr.__set__(self, value)

    def final_values(self):
        """ Returns a dict with all the attribute values.

        This dict can be used to feed :meth:`cubicweb.entity.Entity.cw_set`.
        """
        values = {}
        for attr in self._wsme_attributes:
            if isinstance(attr, wsattr):
                if attr.isfinal and not attr.readonly:
                    value = attr.__get__(self, self.__class__)
                    if value is not wsme.types.Unset:
                        values[attr.rtype] = value
        return values

    def to_entity(self, entity):
        """ Update the entity attributes (not the non-final relations).
        """
        values = self.final_values()
        if values:
            entity.cw_set(**values)


class Any(Base):
    """Complex type to carry any type of entity.

    Automatically used for polymorphic relations targets
    """

    __etype__ = 'Any'

    cw_etype = wsme.types.text

    created_by = wsattr()
    creation_date = wsattr()

    modification_date = wsattr()

    owned_by = wsattr()

    def from_entity(self, entity, keyonly=False, fetch=()):
        self.cw_etype = entity.cw_etype
        super(Any, self).from_entity(
            entity, keyonly=keyonly, fetch=())


class BinaryType(wsme.types.UserType):
    basetype = wsme.types.bytes
    name = 'binary'

    def tobasetype(self, value):
        if value is not None:
            return base64.encodestring(value.getvalue())

    def frombasetype(self, value):
        if value is not None:
            return Binary(base64.decodestring(value))


#: webservice type that map the :class:`cubicweb.Binary` values.
binary = BinaryType()


def scan(vreg, modname):
    """Scan a module for any class inheriting :class:`Base` and register them.
    """
    mod = __import__(modname)
    if '.' in modname:
        for token in modname.split('.')[1:]:
            mod = getattr(mod, token)

    for name, value in inspect.getmembers(mod):
        if inspect.isclass(value) and issubclass(value, Base):
            value.reginit(vreg)



""" 'webservice' controller implementation
"""
import inspect
import itertools
import sys

from cubicweb.web.controller import Controller
from cubicweb.web import NotFound
from cubicweb.predicates import yes
from cubicweb import ValidationError, Forbidden
import cubicweb

import wsme
import wsme.rest.json
import wsme.rest.xml

from wsme.types import text
from cubes.wsme.types import PassThroughType, JsonData, wsattr, Any, Base
from cubes.wsme.predicates import match_ws_etype
from rqlquery.filter import FilterParser

from rqlquery import query
from rqlquery.entity import fill_relation_cache

restformats = {
    'json': wsme.rest.json,
    'xml': wsme.rest.xml
}


def get_fetchable_relations(user, eschema, fetchtree):
    relations = []
    for rschema in eschema.ordered_relations():
        card = eschema.rdef(rschema.type, takefirst=True).cardinality[0]
        if (card in ('*', '+')
                # cardinality '?' with polymorphic relations raises
                # https://www.cubicweb.org/ticket/4482382
                or card == '?' and len(rschema.objects(eschema)) > 1):
            continue

        if not rschema.final and fetchtree.get(rschema.type):
            continue
        if rschema.type in (
                'eid', 'has_text', 'cw_source', 'is'):
            continue
        if rschema.objects()[0].type in ('Password',):
            continue
        if any(not user.matching_groups(es.get_groups('read'))
                for es in rschema.objects(eschema)):
            continue
        relations.append(rschema)
    return relations


def tree_touch_path(tree, path):
    node = tree
    for name in path:
        node = node.setdefault(name, {})


def to_fetchtree(fetchlist):
    fetchtree = {}

    for rel in fetchlist:
        tree_touch_path(fetchtree, rel.split('.'))

    return fetchtree


def prefetch(cnx, entities, fetchtree):
    schema = cnx.vreg.schema
    eschema = schema.eschema(entities[0].cw_etype)

    for rel, subfetch in fetchtree.items():
        if rel == '' or '[' in rel:
            continue
        role = 'object' if rel.startswith('<') else 'subject'
        rtype = rel.strip('<>')

        try:
            rschema = schema.rschema(rtype)
        except KeyError:
            continue
        if rschema.final:
            continue

        tgt_schemas = rschema.targets(eschema, role)
        tgt_schema = tgt_schemas[0] if len(tgt_schemas) == 1 else None

        attrs = [
            attr for attr in subfetch
            if attr == '' or schema.rschema(attr).final
        ]

        fetchall = ('' in attrs or subfetch and not attrs) and tgt_schema
        if fetchall:
            fetchattrs = [
                rs.type for rs in get_fetchable_relations(
                    cnx.user, tgt_schema, subfetch)
            ]
        elif not attrs:
            fetchattrs = ('modification_date',)
        else:
            fetchattrs = attrs

        fill_relation_cache(cnx, entities, rtype, role, fetchattrs=fetchattrs)

        if tgt_schema:
            rel_targets = list(itertools.chain(
                *[e.related(rtype, role, entities=True) for e in entities]))

            if rel_targets:
                prefetch(cnx, rel_targets, subfetch)


class signature(wsme.api.signature):
    def __call__(self, func):
        arg_names = self.options.pop('arg_names', None)
        if arg_names:
            arg_defaults = self.options.pop('arg_defaults', ())
            arg_types = self.arg_types
            self.arg_types = None
        func = super(signature, self).__call__(func)
        if arg_names:
            fd = wsme.api.FunctionDefinition.get(func)
            if fd.body_type is not None:
                arg_types.append(fd.body_type)
            for argname, datatype, default in itertools.izip_longest(
                    arg_names, arg_types, arg_defaults):
                if datatype is wsme.types.HostRequest:
                    fd.pass_request = argname
                else:
                    fd.arguments.append(wsme.api.FunctionArgument(
                        argname, datatype, False, default))
        return func


class ParamsAdapter(dict):
    """ Adapter for passing cubicweb request params to
    :func:`wsme.rest.args.get_args`
    """

    def getall(self, path):
        v = self[path]
        if not isinstance(v, list):
            v = [v]
        return v


class WSController(Controller):
    """
    A controller that rely on WSME to provide webservice API for an entity.
    """
    __regid__ = 'webservice'
    __abstract__ = True

    @classmethod
    def resolve_types(cls, registry):
        """ Late-resolve the types of the function signatures.

        This function is called at regitering time (by :meth:`__registered__`).
        """
        for name, attr in inspect.getmembers(cls, wsme.api.iswsmefunction):
            funcdef = wsme.api.FunctionDefinition.get(attr)
            funcdef.resolve_types(registry)

    @classmethod
    def __registered__(cls, reg):
        cls.resolve_types(reg.vreg.wsme_registry)

    def publish(self, rset):
        """ Main entry-point of the controller.

        Will dispatch the request to the adequate function depending on the
        http method and the form/rset content.

        It also takes care of converting the inputs (form & body) to call
        arguments using the WSME api, based on the function signatures.

        The following `form` values are used, which are normaly set by
        :class:`cubes.wsme.views.RestPathEvaluator`:

        -   `_ws_method`: the HTTP method
        -   `_ws_etype`: the etype (ignored, only used by the selector)
        -   `_ws_rtype`: the relation type if provided
        -   `_ws_rtype_target`: An option relation target id

        If the :arg:`rset` contains an entity, it will be considered as the
        target of the API call.

        """
        content_type = self._cw.get_header('Content-Type')
        accept = self._cw.get_header('Accept')

        restformat = None

        for accept in self._cw.parse_accept_header('Accept'):
            for candidate in restformats.values():
                if accept in candidate.accept_content_types:
                    restformat = candidate
                    break
            if restformat is not None:
                break

        if restformat is None:
            for candidate in restformats.values():
                if self._cw.content_type in candidate.accept_content_types:
                    restformat = candidate
                    break

        if restformat is None:
            restformat = restformats['json']

        try:
            methodname = self._cw.form.pop('_ws_method')
            self._cw.form.pop('_ws_etype')
            func = self._cw.form.pop('_ws_func', None)
            rtype = self._cw.form.pop('_ws_rtype', None)
            rtype_target = self._cw.form.pop('_ws_rtype_target', None)

            prefix, args = '', []
            if rset is not None and rset.rowcount == 1:
                prefix = 'entity_'
                args.append(rset.one())

            if func:
                prefix += 'func_' + func.replace('.', '_') + '_'
                args.append(func)

            if rtype:
                prefix += 'rtype_'
                args.append(rtype)

            if rtype_target:
                prefix += 'target_'
                args.append(rtype_target)

            methodname = prefix + methodname

            method = getattr(self, methodname, None)

            if method is None or not wsme.api.iswsmefunction(method):
                # XXX Raise a 405 http error
                raise NotFound()

            funcdef = wsme.api.FunctionDefinition.get(method)
            if func:
                args.append(funcdef)

            if '_' in self._cw.form:
                del self._cw.form['_']

            self._cw.content.seek(0)
            content = self._cw.content.read()
            self._cw.content.seek(0)
            args, kwargs = wsme.rest.args.get_args(
                funcdef, args, {}, ParamsAdapter(self._cw.form), None,
                content, content_type
            )

            self._cw.set_content_type(restformat.content_type)
            result = method(*args, **kwargs)
            if isinstance(result, wsme.api.Response):
                self._cw.status_out = result.status_code
                result = result.obj
            return restformat.encode_result(result, funcdef.return_type)
        except NotFound:
            self._cw.status_out = 404
            return restformat.encode_error(None, {
                'faultcode': 'Client',
                'faultstring': 'Not Found'
            })
        except ValidationError, e:
            fmtexc = {
                'faultcode': 'Client',
                'faultstring': "ValidationError",
                'errors': e.errors
            }
            self._cw.status_out = 400
            return restformat.encode_error(funcdef, fmtexc)
        except:
            self.exception('Error while calling method {}'.format(methodname))
            self._cw.cnx.rollback()
            exc_info = sys.exc_info()
            fmtexc = wsme.api.format_exception(
                exc_info, debug=self._cw.vreg.config.debugmode)
            if isinstance(exc_info[1], cubicweb.Unauthorized):
                fmtexc['faultcode'] = 'Client'
                self._cw.status_out = 401
            elif fmtexc['faultcode'] == 'Client':
                self._cw.status_out = 400
            elif fmtexc['faultcode'] == 'Server':
                self._cw.status_out = 500
            return restformat.encode_error(funcdef, fmtexc)


class WSCRUDController(WSController):
    """
    An entity type CRUD controller

    The displatch is summarized in this table, where 'entity' means that an
    entity exists in the rset:

    .. csv-table::

        form-rset / verb, GET, POST, PUT, DELETE
        , :meth:`_get`, :meth:`_post`, ,
        entity, :meth:`_entity_get`, , :meth:`_entity_put`, :meth:`_entity_delete`
        "entity, _ws_rtype", :meth:`_entity_rtype_get`, :meth:`_entity_rtype_post`
        "entity, _ws_rtype, _ws_rtype_target", , , , :meth:`_entity_rtype_target_delete`
    """

    #: The entity type
    __cwetype__ = None
    __select__ = yes()

    allowed_functions = {
        'IWorkflowable.fire_transition': (
            ('TrInfo', text), ('tr', 'comment', 'commentformat'))
    }

    def _get_entity(self, data):
        """
        Get an entity and update/create it and its related entities all along.

        :param data: A webservice type instance
        """
        eid, values, relation_values = self._handle_data(data)
        if eid:
            entity = self._cw.entity_from_eid(eid)
            if values:
                entity.cw_set(**values)
        else:
            entity = self._cw.create_entity(data.__etype__, **values)
        for (rtype, role, inlined), targets in relation_values.items():
            if role == 'subject' and inlined:
                entity.cw_set(**{rtype: targets})
            if not isinstance(targets, list):
                if targets is None:
                    targets = ()
                else:
                    targets = [targets]
            d = {'x': entity.eid}
            if role == 'subject':
                relation = 'X %s Y' % rtype
            else:
                relation = 'Y %s X' % rtype
            if not targets:
                entity._cw.execute(
                    'DELETE %s WHERE X eid %%(x)s' % relation, d)
            else:
                d.update({'y%s' % i: t.eid for i, t in enumerate(targets)})
                entity._cw.execute(
                    "SET %(relation)s WHERE NOT %(relation)s, X eid %%(x)s, "
                    "Y eid IN (%(list)s)" % {
                        'relation': relation,
                        'list': ', '.join(
                            '%%(y%s)s' % x for x in range(len(targets))
                        )},
                    d)
                entity._cw.execute(
                    "DELETE %(relation)s WHERE %(relation)s, X eid %%(x)s, "
                    "NOT Y eid IN (%(list)s)" % {
                        'relation': relation,
                        'list': ', '.join(
                            '%%(y%s)s' % x for x in range(len(targets))
                        )},
                    d)
        return entity

    def _get_entities(self, datalist):
        """
        Get a list of entities from a list a webservice data
        """
        return [self._get_entity(data) for data in datalist]

    def _handle_data(self, data):
        """
        Handle webservice data.

        It returns a tuple `(eid, values, relation_values)`, where eid can be
        None if the data had none, `values` contains the final and inlined
        values, and `relation_values` the relation values. These variables are
        dictionnaries that can be fed cw_set().

        While handling the entity data, the related entities present in the
        data will be updated/create (via :meth:`_get_entity`).
        """
        eid = data.eid if data.eid else None
        values = {}
        relation_values = {}
        for attr in data._wsme_attributes:
            if not isinstance(attr, wsattr):
                continue
            if attr.rtype == 'eid':
                continue
            value = attr.__get__(data, data.__class__)
            if value is wsme.types.Unset:
                continue
            if attr.isfinal:
                values[attr.rtype] = value
            else:
                if value is not None:
                    if wsme.types.isarray(attr.datatype):
                        value = self._get_entities(value)
                    else:
                        value = self._get_entity(value)
                if attr.inlined:
                    values[attr.rtype] = value
                else:
                    relation_values[
                        (attr.rtype, attr.role, attr.inlined)] = value
        return eid, values, relation_values

    def _update(self, data):
        """ Update an existing entity from ws data"""
        assert data.eid, "missing eid on data"
        return self._get_entity(data)

    def _create(self, data):
        """ Create an entity from ws data"""
        if data.eid:
            raise ValueError(
                "Cannot create with a fixed eid. Please remove it")
        return self._get_entity(data)

    def _call_func(self, _entity, _function_name, _funcdef, **kwargs):
        iname, fname = _function_name.split('.')
        if _function_name not in self.allowed_functions:
            raise Forbidden()
        adapter = _entity.cw_adapt_to(iname)
        f = getattr(adapter, fname)
        result = f(**kwargs)
        self._cw.cnx.commit()
        if inspect.isclass(_funcdef.return_type) and \
                issubclass(_funcdef.return_type, Base) and \
                not isinstance(result, _funcdef.return_type):
            result = _funcdef.return_type(result)
        return result

    @classmethod
    def __registered__(cls, reg):
        if not hasattr(cls, 'get'):
            cls.get = wsme.api.signature(
                [cls.__cwetype__], [text], JsonData, int, int, [text],
                bool, wrap=True
            )(cls._get)

        if not hasattr(cls, 'post'):
            cls.post = wsme.api.signature(
                cls.__cwetype__, [text], bool, body=cls.__cwetype__, wrap=True
            )(cls._post)

        if not hasattr(cls, 'entity_get'):
            cls.entity_get = wsme.api.signature(
                cls.__cwetype__, PassThroughType, [text], wrap=True
            )(cls._entity_get)

        if not hasattr(cls, 'entity_put'):
            cls.entity_put = wsme.api.signature(
                cls.__cwetype__, PassThroughType, [text], body=cls.__cwetype__,
                wrap=True
            )(cls._entity_put)

        if not hasattr(cls, 'entity_delete'):
            cls.entity_delete = wsme.api.signature(
                None, PassThroughType, [text], wrap=True
            )(cls._entity_delete)

        if not hasattr(cls, 'entity_rtype_post'):
            cls.entity_rtype_post = wsme.api.signature(
                None, PassThroughType, text, body=int, wrap=True
            )(cls._entity_rtype_post)

        if not hasattr(cls, 'entity_rtype_get'):
            cls.entity_rtype_get = wsme.api.signature(
                [Any], PassThroughType, text, [text], int, int,
                bool, wrap=True
            )(cls._entity_rtype_get)

        if not hasattr(cls, 'entity_rtype_target_delete'):
            cls.entity_rtype_target_delete = wsme.api.signature(
                None, PassThroughType, text, int, wrap=True
            )(cls._entity_rtype_target_delete)

        for name, (argtypes, argnames) in cls.allowed_functions.items():
            if len(argtypes) == 1:
                method = 'get'
            else:
                method = 'post'
            # The function name will be passed
            argtypes = (
                (argtypes[0],
                 PassThroughType, PassThroughType, PassThroughType)
                + argtypes[1:])
            argnames = ('_entity', '_function_name', '_funcdef',) + argnames
            mname = 'entity_func_%s_%s' % (
                name.replace('.', '_'), method)
            setattr(cls, mname, signature(
                *argtypes, wrap=True, arg_names=argnames)(cls._call_func))

        if cls.__select__ is WSCRUDController.__select__:
            cls.__select__ = match_ws_etype(cls.__cwetype__.__name__)

        super(WSCRUDController, cls).__registered__(reg)

    def _get(self, orderby=None, filter=None, limit=0, offset=0, fetch=[],
             keyonly=False):
        """List entities with an optional filter.

        Default implementation of `GET /etype`.

        :param filter:
        :param fetch: A list of relations and subrelations of which the target
                      entities will be returned.
        """
        fetchtree = to_fetchtree(fetch)

        q = query.Query(self._cw.vreg.schema, self.__cwetype__.__etype__)

        if not keyonly:
            eschema = self._cw.vreg.schema.eschema(self.__cwetype__.__etype__)
            cols = [
                rs.type for rs in get_fetchable_relations(
                    self._cw.user, eschema, fetchtree)
            ]
            q = q.add_column(*cols)

        if orderby:
            q = q.orderby(*orderby)
        if filter:
            q = q.filter(FilterParser(
                self._cw.vreg.schema, self.__cwetype__.__etype__, filter
            ).parse())
        if limit:
            q = q.limit(limit)
        if offset:
            q = q.offset(offset)
        entities = list(q.all(self._cw))
        if entities:
            prefetch(self._cw, entities, fetchtree)
        return [
            self.__cwetype__(e, keyonly=keyonly, fetch=fetch)
            for e in entities
        ]

    def _post(self, fetch=[], keyonly=False, data=None):
        """ Default implementation of `POST /etype`."""
        try:
            entity = self._create(data)
        except:
            self._cw.cnx.rollback()
            raise
        else:
            # XXX We dont really want a commit, just make sure all hooks and
            # operations are done.
            self._cw.cnx.commit()
        return self.__cwetype__(entity, keyonly=keyonly, fetch=fetch)

    def _entity_get(self, entity, fetch=[]):
        """ Default implementation of `GET /etype/eid`."""
        return self.__cwetype__(entity, fetch=fetch)

    def _entity_put(self, entity, fetch=[], data=None):
        """ Default implementation of `PUT /etype/eid`."""
        if not data.eid:
            data.eid = entity.eid
        entity = self._update(data)
        # XXX We dont really want a commit, just make sure all hooks and
        # operations are done.
        self._cw.cnx.commit()
        # XXX We should clear the cache of entity to really have the db data if
        # they were modified by some hooks
        return self.__cwetype__(entity, fetch=fetch)

    def _entity_delete(self, entity):
        """ Default implementation of `DELETE /etype/eid`."""
        entity.cw_delete()
        return wsme.api.Response(None, 204)

    def _entity_rtype_post(self, entity, rtype, eid):
        """ Default implementation of `POST /etype/eid/rtype`."""
        if rtype.startswith('<'):
            rtype = 'reverse_' + rtype[1:]
        entity.cw_set(**{rtype: eid})

    def _entity_rtype_get(self, entity, rtype, orderby=None, limit=None,
                          offset=None, keyonly=False):
        """ Default implementation of `GET /etype/eid/rtype`."""
        if rtype.startswith('<'):
            rtype, role = rtype[1:], 'object'
        else:
            role = 'subject'
        rql = entity.cw_related_rql(rtype, role, limit=limit)
        if offset:
            rql = rql.replace(
                'LIMIT %s' % limit,
                'LIMIT %s OFFSET %s' % (limit, offset))
        if orderby:
            raise NotImplementedError('yet')
        return [
            Any(e, keyonly=keyonly) for e in self._cw.execute(
                rql, {'x': entity.eid}).entities()
        ]

    def _entity_rtype_target_delete(self, entity, rtype, eid):
        """ Default implementation of `DELETE /etype/eid/rtype/eid`."""
        if rtype.startswith('<'):
            rtype = rtype[1:]
            relation = 'X %s E'
        else:
            relation = 'E %s X'
        relation = relation % rtype
        rql = "DELETE " + relation + " WHERE X eid %(x)s, E eid %(e)s"
        self._cw.execute(rql, {'e': entity.eid, 'x': eid})
        return wsme.api.Response(None, 204)



# pylint: disable=W0622
"""cubicweb-wsme application packaging information"""

modname = 'wsme'
distname = 'cubicweb-wsme'

numversion = (0, 2, 0)
version = '.'.join(str(num) for num in numversion)

license = 'LGPL'
author = 'Christophe de Vienne'
author_email = 'christophe@unlish.com'
description = 'Easily build a webservice API on top of a cubic web database"'
web = 'http://www.cubicweb.org/project/%s' % distname

__depends__ = {
    'cubicweb': '>= 3.19.0',
    'wsme': None,
    'rqlquery': '>= 0.3.0'
}

__recommends__ = {}

classifiers = [
    'Environment :: Web Environment',
    'Framework :: CubicWeb',
    'Programming Language :: Python',
    'Programming Language :: JavaScript',
    ]

from os import listdir as _listdir
from os.path import join, isdir
from glob import glob

THIS_CUBE_DIR = join('share', 'cubicweb', 'cubes', modname)


def listdir(dirpath):
    return [join(dirpath, fname) for fname in _listdir(dirpath)
            if fname[0] != '.' and not fname.endswith('.pyc')
            and not fname.endswith('~')
            and not isdir(join(dirpath, fname))]

data_files = [
    # common files
    [THIS_CUBE_DIR, [fname for fname in glob('*.py') if fname != 'setup.py']],
    ]
# check for possible extended cube layout
for dname in (
        'entities', 'views', 'sobjects', 'hooks', 'schema', 'data', 'wdoc',
        'i18n', 'migration'):
    if isdir(dname):
        data_files.append([join(THIS_CUBE_DIR, dname), listdir(dname)])
# Note: here, you'll need to add subdirectories if you want
# them to be included in the debian package
